"""Provides an implementation of various standard templates found when using topological quantum
error correction.
"""

from .fixed_boundary import (
    FixedBoundaryConventionGenerator as FixedBoundaryConventionGenerator,
)
from .fixed_bulk import FixedBulkConventionGenerator as FixedBulkConventionGenerator
