"""Contains exceptions related to :mod:`tqec.visualisation`."""

from tqec.utils.exceptions import TQECError


class TQECDrawingError(TQECError):
    """Exception raised when an error is encountered during drawing."""

    pass
