"""Defines the data structures to represent the logical computations.

This module provides high-level abstractions to represent the fault-tolerant logical computations
protected by surface code.

"""
