"""Defines several helper methods to facilitate visualising QEC circuit simulation."""

from .inset import add_inset_axes3d as add_inset_axes3d
from .inset import plot_observable_as_inset as plot_observable_as_inset
