"""Defines several helper methods to facilitate QEC circuit simulation."""

from .plotting import plot_observable_as_inset as plot_observable_as_inset
from .simulation import start_simulation_using_sinter as start_simulation_using_sinter
from .split import heuristic_custom_error_key as heuristic_custom_error_key
from .split import split_stats_for_observables as split_stats_for_observables
